from __future__ import annotations

import datetime
import typing as tp

import jijmodeling as jm
import pandas as pd
from typing_extensions import TypeAlias

if tp.TYPE_CHECKING:
    from jijbench.containers.containers import Artifact, Record, Table
    from jijbench.experiment.experiment import Experiment
    from jijbench.node.base import DataNode

T = tp.TypeVar("T")


# node
DataNodeT = tp.TypeVar("DataNodeT", bound="DataNode[tp.Any]")
DataNodeInT = tp.TypeVar("DataNodeInT", bound="DataNode[tp.Any]")
DataNodeOutT = tp.TypeVar("DataNodeOutT", bound="DataNode[tp.Any]")


# element
DateTypes: TypeAlias = tp.Union[str, datetime.datetime, pd.Timestamp]
NumberTypes: TypeAlias = tp.Union[int, float]

# solver
ModelType: TypeAlias = tp.Tuple[jm.Problem, jm.PH_VALUES_INTERFACE]


# mapping
ContainerT = tp.TypeVar("ContainerT", "Artifact", "Experiment", "Record", "Table")
ContainerTypes: TypeAlias = tp.Union["Artifact", "Experiment", "Record", "Table"]
ContainerListTypes: TypeAlias = tp.Union[
    tp.List["Artifact"], tp.List["Experiment"], tp.List["Record"], tp.List["Table"]
]
ArtifactDataType: TypeAlias = tp.Dict[
    tp.Hashable, tp.Dict[tp.Hashable, "DataNode[tp.Any]"]
]
ExperimentDataType: TypeAlias = tp.Tuple["Artifact", "Table"]



from __future__ import annotations

import abc
import pathlib
import typing as tp
import uuid
import warnings
from dataclasses import dataclass, field

import pandas as pd

from jijbench.consts.path import DEFAULT_RESULT_DIR
from jijbench.containers.containers import Artifact, Container, Record, Table
from jijbench.elements.base import Callable
from jijbench.elements.id import ID
from jijbench.functions.concat import Concat
from jijbench.functions.factory import ArtifactFactory, TableFactory
from jijbench.io.io import save
from jijbench.solver.base import Parameter, Response
from jijbench.typing import ArtifactDataType, ExperimentDataType


@dataclass
class Experiment(Container[ExperimentDataType]):
    data: tuple[Artifact, Table] = field(default_factory=lambda: (Artifact(), Table()))
    name: str = field(default_factory=lambda: str(uuid.uuid4()))
    autosave: bool = field(default=True, repr=False)
    savedir: str | pathlib.Path = field(default=DEFAULT_RESULT_DIR, repr=False)

    def __post_init__(self):
        super().__post_init__()

        if not isinstance(self.data, tuple):
            raise TypeError(f"Data must be a tuple, got {type(self.data)}.")

        if len(self.data) != 2:
            raise ValueError(f"Data must be a tuple of length 2, got {len(self.data)}.")

        if not isinstance(self.data[0], Artifact):
            raise TypeError(
                f"First element of data must be an Artifact, got {type(self.data[0])}."
            )

        if not isinstance(self.data[1], Table):
            raise TypeError(
                f"Second element of data must be a Table, got {type(self.data[1])}."
            )

        if self.data[0].name is None:
            self.data[0].name = self.name

        if self.data[1].name is None:
            self.data[1].name = self.name

        self.savedir = pathlib.Path(self.savedir)
        setattr(self, "state", _Created(self.name))

    def __len__(self) -> int:
        """
        Perform the operation __len__.
        """
        a_len = len(self.data[0])
        t_len = len(self.data[1])
        if a_len != t_len:
            warnings.warn(
                f"The length of artifact object and table object are different: {a_len} != {t_len}, return the larger one."
            )
        return max(a_len, t_len)

    def __enter__(self) -> Experiment:
        """Set up Experiment.
        Automatically makes a directory for saving the experiment, if it doesn't exist.
        """

        setattr(self, "state", _Running(self.name))
        pathlib.Path(self.savedir).mkdir(parents=True, exist_ok=True)
        return self

    def __exit__(self, exception_type, exception_value, traceback) -> None:
        """Saves the experiment if autosave is True."""
        state = getattr(self, "state")

        if self.autosave:
            state.save(self)

        setattr(self, "state", _Waiting(self.name))

    @property
    def artifact(self) -> ArtifactDataType:
        """Return the artifact of the experiment as a dictionary."""
        return self.data[0].view()

    @property
    def table(self) -> pd.DataFrame:
        """Return the table of the experiment as a pandas dataframe."""
        return self.data[1].view()

    @property
    def params_table(self) -> pd.DataFrame:
        """Return the parameters table of the experiment as a pandas dataframe."""
        bools = self.data[1].data.applymap(lambda x: isinstance(x, Parameter))
        return self.table[bools].dropna(axis=1)

    @property
    def solver_table(self) -> pd.DataFrame:
        """Return the solver table of the experiment as a pandas dataframe."""

        bools = self.data[1].data.applymap(lambda x: isinstance(x, Callable))
        return self.table[bools].dropna(axis=1)

    @property
    def response_table(self) -> pd.DataFrame:
        """Return the returns table of the experiment as a pandas dataframe."""
        bools = self.data[1].data.apply(lambda x: not isinstance(x[0], Response))
        droped_columns = self.data[1].data.columns[bools]
        return self.table.drop(columns=droped_columns)

    @classmethod
    def validate_data(cls, data: ExperimentDataType) -> ExperimentDataType:
        """Validate the data of the experiment.

        Args:
            data (ExperimentDataType): The data to validate.

        Raises:
            TypeError: If data is not an instance of ExperimentDataType or if the first element of data is not an instance of Artifact or if the second element of data is not an instance of Table.

        Returns:
            ExperimentDataType: The validated data.
        """
        artifact, table = data
        if not isinstance(artifact, Artifact):
            raise TypeError(
                f"Type of attribute data is {ExperimentDataType}, and data[0] must be Artifact instead of {type(artifact).__name__}."
            )
        if not isinstance(table, Table):
            raise TypeError(
                f"Type of attribute data is {ExperimentDataType}, and data[1] must be Table instead of {type(artifact).__name__}."
            )
        return data

    def view(self) -> tuple[dict, pd.DataFrame]:
        """Return a tuple of the artifact dictionary and table dataframe."""
        return (self.data[0].view(), self.data[1].view())

    def append(self, record: Record) -> None:
        """Append a new record to the experiment.

        Args:
            record (Record): The record to be appended to the experiment.
        """
        state = getattr(self, "state")
        state.append(self, record)

    def save(self):
        """Save the experiment."""
        pathlib.Path(self.savedir).mkdir(parents=True, exist_ok=True)
        state = getattr(self, "state")
        state.save(self)


class _ExperimentState(metaclass=abc.ABCMeta):
    def __init__(self, name: tp.Hashable) -> None:
        self.name = name

    @abc.abstractmethod
    def append(self, context: Experiment, record: Record) -> None:
        pass

    @abc.abstractmethod
    def save(self) -> None:
        pass


class _Created(_ExperimentState):
    def append(self, context: Experiment, record: Record) -> None:
        record.name = len(context)
        _append(context, record)
        context.state = _Waiting(self.name)

    def save(self, context: Experiment) -> None:
        save(context, savedir=context.savedir)


class _Waiting(_ExperimentState):
    def append(self, context: Experiment, record: Record) -> None:
        record.name = len(context)
        _append(context, record)

    def save(self, context) -> None:
        save(context, savedir=context.savedir)


class _Running(_ExperimentState):
    def append(self, context: Experiment, record: Record) -> None:
        run_id = ID().data
        if isinstance(self.name, str):
            record.name = (self.name, run_id)
        elif isinstance(self.name, tp.Iterable):
            record.name = (*self.name, run_id)
        else:
            record.name = (self.name, run_id)
        _append(context, record)

    def save(
        self,
        context: Experiment,
    ) -> None:
        a, t = context.data
        latest = t.index[-1]

        ai = Artifact({latest: a.data[latest]}, a.name)
        ti = Table(t.data.loc[[latest]], t.name)

        experiment = Experiment(
            (ai, ti), context.name, autosave=context.autosave, savedir=context.savedir
        )
        save(experiment, savedir=context.savedir, mode="a")


def _append(context: Experiment, record: Record) -> None:
    concat: Concat[Experiment] = Concat()
    data = (ArtifactFactory()([record]), TableFactory()([record]))
    other = type(context)(
        data, context.name, autosave=context.autosave, savedir=context.savedir
    )
    node = context.apply(
        concat,
        [other],
        name=context.name,
        autosave=context.autosave,
        savedir=context.savedir,
    )
    context._init_attrs(node)



from __future__ import annotations

import numpy as np
import pandas as pd

from jijbench.containers.containers import Artifact, Table
from jijbench.elements.array import Array
from jijbench.experiment.experiment import Experiment
from jijbench.functions.concat import Concat
from jijbench.functions.factory import RecordFactory
from jijbench.functions.metrics import (
    FeasibleRate,
    ResidualEnergy,
    SuccessProbability,
    TimeToSolution,
)
from jijbench.node.base import FunctionNode
from jijbench.solver.jijzept import SampleSet


class Evaluation(FunctionNode[Experiment, Experiment]):
    """Evaluate the benchmark results."""

    def __call__(
        self,
        inputs: list[Experiment],
        opt_value: float | None = None,
        pr: float = 0.99,
    ) -> Experiment:
        return super().__call__(inputs, opt_value=opt_value or np.nan, pr=pr)

    def operate(
        self,
        inputs: list[Experiment],
        opt_value: float,
        pr: float = 0.99,
    ) -> Experiment:
        """Calculate the typincal metrics of benchmark results.

        The metrics are as follows:
            - success_probability: Solution that is feasible and less than or equal to opt_value is counted as success, which is NaN if `opt_value` is not given.
            - feasible_rate: Rate of feasible solutions out of all solutions.
            - residual_energy: Difference between average objective of feasible solutions and `opt_value`, which is NaN if `opt_value` is not given.
            - TTS(optimal): Time to obtain opt_value with probability `pr`, which is NaN if opt_value is not given.
            - TTS(feasible): Time to obtain feasible solutions with probability `pr`.
            - TTS(derived): Time to obtain minimum objective among feasible solutions with probability `pr`.

        Args:
            opt_value (float, optional): Optimal value for instance_data.
            pr (float, optional): Probability of obtaining optimal value. Defaults to 0.99.

        Returns:
            Experiment: Experiment object included evalution results.
        """

        def f(x: pd.Series, opt_value: float, pr: float) -> pd.Series:
            inputs: list[SampleSet] = x.tolist()
            node = Concat()(inputs)
            arrays = [
                Array(v, k)
                for k, v in Table._extract(node.data).items()
                if isinstance(v, np.ndarray) and k != "num_occurrences"
            ]
            metrics = [
                SuccessProbability()([node], opt_value=opt_value),
                FeasibleRate()([node]),
                ResidualEnergy()([node], opt_value=opt_value),
                TimeToSolution()([node], pr=pr, opt_value=opt_value, base="optimal"),
                TimeToSolution()([node], pr=pr, base="feasible"),
                TimeToSolution()([node], pr=pr, base="derived"),
            ]
            metrics += sum(
                [
                    [array.min(), array.max(), array.mean(), array.std()]
                    for array in arrays
                ],
                [],
            )
            record = RecordFactory()(metrics)
            return record.data

        experiment = Concat()(inputs)

        artifact, table = experiment.data

        sampleset_columns = [
            col
            for col in table.columns
            if all(map(lambda x: isinstance(x, SampleSet), table.data[col]))
        ]
        data = table.data[sampleset_columns].apply(
            f, opt_value=opt_value, pr=pr, axis=1
        )

        metrics_artifact = Artifact(data.to_dict("index"))
        artifact = Concat()([artifact, metrics_artifact])

        metrics_table = Table(data)
        table = Concat()([table, metrics_table], axis=1)

        experiment.data = (artifact, table)
        return experiment



from __future__ import annotations

from collections import OrderedDict

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt

from jijbench.visualization.figure.interface import Figure

# TODO: docstringの追加


class Schedule(Figure):
    """Visualize schedule.

    Attributes:
        data (OrderedDict):the dict of schedule data. the key is task label, and the value is tuple of three lists. Each is the list of worker, the start time of work, the time length of work.
        fig_ax (tuple[matplotlib.figure.Figure, matplotlib.axes.Subplot]): Figure and Axes of matplotlib. Available after show method is called.
    Example:
        The code below plots a work shift schedule,
        where worker1 works on task0 for 5 unit time from time 3
        and worker2 works on task0 for 6 unit time from time 4.
        The style of the graph (e.g. color) can be changed by arguments of the `.show` method.

        ```python
        >>> from jijbench.figure.schedule import Schedule

        >>> schedule = Schedule()
        >>> schedule.add_data(task_label="task0",
        >>>     workers=[1, 2],
        >>>     start_times=[3, 4],
        >>>     time_lengths=[5, 6],
        >>> )

        >>> schedule.show(color_list=["red"])
        ```
    """

    def __init__(
        self,
    ) -> None:
        self.data = OrderedDict()
        self._fig_ax = None
        self._workers_set = set()

    def add_data(
        self,
        task_label: str,
        workers: list[int] | npt.NDArray,
        start_times: list[int | float] | npt.NDArray,
        time_lengths: list[int | float] | npt.NDArray,
    ) -> None:
        """Add schedule data to data attribute for plot.

        Args:
            task_label (str): the label of the task.
            workers (list[int] | npt.NDArray): the list of worker indices. the length of the list is the number of tasks belonging to task_label.
            start_times (list[int | float] | npt.NDArray): the list of the start time of work. the length of the list is the number of tasks belonging to task_label.
            time_lengths (list[int | float] | npt.NDArray): the list of the time length of work. the length of the list is the number of tasks belonging to task_label.
        """
        if isinstance(workers, np.ndarray):
            workers = workers.tolist()
        if isinstance(start_times, np.ndarray):
            start_times = start_times.tolist()
        if isinstance(time_lengths, np.ndarray):
            time_lengths = time_lengths.tolist()

        if len(workers) != len(start_times):
            raise ValueError("workers and start_times must be the same length.")
        if len(workers) != len(time_lengths):
            raise ValueError("workers and time_lengths must be the same length.")
        self.data.update([(task_label, (workers, start_times, time_lengths))])
        self._workers_set |= set(workers)

    def show(
        self,
        figsize: tuple[int | float] | None = None,
        title: str | None = None,
        color_list: list | None = None,
        alpha_list: list[float] | None = None,
        xlabel: str | None = None,
        ylabel: str | None = None,
        xticks: list[int | float] | None = None,
        yticks: list[int | float] | None = None,
    ):
        """Plot schedule data which you passed to the add_data method.

        The arguments of the show method are passed to the plot of matplotlib.

        Args:
            figsize (tuple[int | float] | None): the size of figure. The default uses matplotlib's default value.
            title (str | None): the title of figure. Defaults to "time series".
            color_list (list | None): the list of plot line color. The default uses matplotlib's default value.
            alpha_list (list[float] | None): the list of plot line transparency. The default is 1.0 for each plot line.
            xlabel (str | None): the xlabel of figure. Defaults to None.
            ylabel (str | None): the ylabel of figure. Defaults to None.
            xticks (list[int | float] | None): the xticks of figure. The default uses matplotlib's default.
            yticks (list[int | float] | None): the yticks of figure. The default uses matplotlib's default.
        """

        data = self.data

        if len(data) == 0:
            raise RuntimeError(
                "no plot data. Add at least one plot data with add_data method."
            )

        if title is None:
            title = "schedule"

        if (color_list is not None) and (len(color_list) != len(data)):
            raise ValueError("color_list and data must be same length.")

        if alpha_list is None:
            alpha_list = [0.5] * len(data)
        elif len(alpha_list) != len(data):
            raise ValueError("alpha_list and data must be same length.")

        if xlabel is None:
            xlabel = "time"

        if ylabel is None:
            ylabel = "worker"

        if yticks is None:
            yticks = self.workers

        fig, ax = plt.subplots(figsize=figsize)
        fig.suptitle(title)

        for i, (task_label, plot_data) in enumerate(data.items()):
            workers, start_times, time_lengths = plot_data
            color = None if color_list is None else color_list[i]
            ax.barh(
                workers,
                time_lengths,
                left=start_times,
                height=0.5,
                label=task_label,
                color=color,
                alpha=alpha_list[i],
            )
            centers = np.array(start_times) + np.array(time_lengths) / 2

            for battery, center, time_length in zip(workers, centers, time_lengths):
                ax.text(
                    center,
                    battery,
                    str(int(time_length)),
                    ha="center",
                    va="center",
                )

        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        if xticks is not None:
            ax.set_xticks(xticks)
        ax.set_yticks(yticks)

        ax.legend(
            ncol=len(data),
            bbox_to_anchor=(0, 0.99),
            loc="lower left",
        )

        self._fig_ax = (fig, ax)

    @property
    def fig_ax(self):
        if self._fig_ax is None:
            raise AttributeError(
                "fig_ax attribute is available after show method is called."
            )
        else:
            return self._fig_ax

    @property
    def workers(self):
        return sorted(self._workers_set)



from __future__ import annotations

import pathlib
import typing as tp

import dill

from jijbench.consts.path import DEFAULT_RESULT_DIR
from jijbench.containers.containers import Artifact, Table
from jijbench.elements.base import Any
from jijbench.functions.concat import Concat

if tp.TYPE_CHECKING:
    from jijbench.experiment.experiment import Experiment


@tp.overload
def save(
    obj: Experiment,
    *,
    savedir: str | pathlib.Path = DEFAULT_RESULT_DIR,
    mode: tp.Literal["w", "a"] = "w",
    index_col: int | list[int] | None = None,
) -> None:
    ...


@tp.overload
def save(
    obj: Artifact,
    *,
    savedir: str | pathlib.Path = DEFAULT_RESULT_DIR,
    mode: tp.Literal["w", "a"] = "w",
) -> None:
    ...


@tp.overload
def save(
    obj: Table,
    *,
    savedir: str | pathlib.Path = DEFAULT_RESULT_DIR,
    mode: tp.Literal["w", "a"] = "w",
    index_col: int | list[int] | None = None,
) -> None:
    ...


def save(
    obj: Artifact | Experiment | Table,
    *,
    savedir: str | pathlib.Path = DEFAULT_RESULT_DIR,
    mode: tp.Literal["w", "a"] = "w",
    index_col: int | list[int] | None = None,
) -> None:
    """Save the given `Artifact`, `Experiment`, or `Table` object.

    Args:
        obj (Artifact | Experiment | Table): The object to be saved.
        savedir (str | pathlib.Path, optional): The directory where the object will be saved. Defaults to DEFAULT_RESULT_DIR.
        mode (Literal[&quot;w&quot;, &quot;a&quot;], optional): The write mode for the file. Must be 'w' or 'a'. Defaults to "w".
        index_col (int | list[int] | None, optional): Index column(s) to set as index while saving the table. Defaults to None.

    Raises:
        ValueError: If the mode is not 'w' or 'a'.
        FileNotFoundError: If the savedir does not exist.
        IOError: If the object is not dillable.
        TypeError: If the object is not an `Artifact`, `Experiment`, or `Table`.
    """
    from jijbench.experiment.experiment import Experiment

    def is_dillable(obj: tp.Any) -> bool:
        try:
            dill.dumps(obj)
            return True
        except Exception:
            return False

    def to_dillable(obj: Artifact | Table) -> Artifact | Table:
        if isinstance(obj, Artifact):
            data = {}
            for k, v in obj.data.items():
                data[k] = {}
                for name, node in v.items():
                    if is_dillable(node.data):
                        data[k].update({name: node})
                    else:
                        data[k].update({name: Any(str(node.data), node.name)})
            return Artifact(data, obj.name)
        else:
            data = obj.data.applymap(
                lambda x: x if is_dillable(x) else Any(str(x.data), x.name)
            )
            return Table(data, obj.name)

    if mode not in ["a", "w"]:
        raise ValueError("Argument mode must be 'a' or 'w'.")

    savedir = savedir if isinstance(savedir, pathlib.Path) else pathlib.Path(savedir)
    if not savedir.exists():
        raise FileNotFoundError(f"Directory {savedir} is not found.")

    if isinstance(obj, Artifact):
        p = savedir / "artifact.dill"
        concat_a: Concat[Artifact] = Concat()

        obj = to_dillable(obj)
        if mode == "a":
            if p.exists():
                obj = concat_a(
                    [
                        load(
                            savedir,
                            return_type="Artifact",
                        ),
                        obj,
                    ]
                )

        with open(p, "wb") as f:
            dill.dump(obj, f)

    elif isinstance(obj, Experiment):
        savedir_a = savedir / obj.name / "artifact"
        savedir_t = savedir / obj.name / "table"
        savedir_a.mkdir(parents=True, exist_ok=True)
        savedir_t.mkdir(parents=True, exist_ok=True)
        save(
            obj.data[0],
            savedir=savedir_a,
            mode=mode,
        )
        save(
            obj.data[1],
            savedir=savedir_t,
            mode=mode,
            index_col=index_col,
        )
    elif isinstance(obj, Table):
        p_csv = savedir / "table.csv"
        p_dill = savedir / "table.dill"
        p_meta = savedir / "meta.dill"
        concat_t: Concat[Table] = Concat()

        obj = to_dillable(obj)
        if mode == "a":
            if p_csv.exists() and p_meta.exists():
                obj = concat_t(
                    [
                        load(
                            savedir,
                            return_type="Table",
                            index_col=index_col,
                        ),
                        obj,
                    ]
                )
        obj.view().to_csv(p_csv)
        meta = {
            "dtype": obj.data.iloc[0].apply(lambda x: x.__class__).to_dict(),
            "name": obj.data.applymap(lambda x: x.name).to_dict(),
            "index": obj.data.index,
            "columns": obj.data.columns,
        }
        with open(p_dill, "wb") as f:
            dill.dump(obj, f)
        with open(p_meta, "wb") as f:
            dill.dump(meta, f)
    else:
        raise TypeError(
            f"Cannnot save type {obj.__class__}. Type of obj must be Artifact or Experiment or Table."
        )


@tp.overload
def load(
    name_or_dir: str | pathlib.Path,
    *,
    experiment_names: list[str] | None = None,
    savedir: str | pathlib.Path = DEFAULT_RESULT_DIR,
    index_col: int | list[int] | None = None,
) -> Experiment:
    ...


@tp.overload
def load(
    name_or_dir: str | pathlib.Path,
    *,
    savedir: str | pathlib.Path = DEFAULT_RESULT_DIR,
    return_type: tp.Literal["Artifact"] = ...,
) -> Artifact:
    ...


@tp.overload
def load(
    name_or_dir: str | pathlib.Path,
    *,
    savedir: str | pathlib.Path = DEFAULT_RESULT_DIR,
    return_type: tp.Literal["Table"] = ...,
    index_col: int | list[int] | None = None,
) -> Table:
    ...


def load(
    name_or_dir: str | pathlib.Path,
    *,
    experiment_names: list[str] | None = None,
    savedir: str | pathlib.Path = DEFAULT_RESULT_DIR,
    return_type: tp.Literal["Artifact", "Experiment", "Table"] = "Experiment",
    index_col: int | list[int] | None = None,
) -> Experiment | Artifact | Table:
    """Load and return an artifact, experiment, or table from the given directory.

    Args:
        name_or_dir (str | pathlib.Path): Name or directory of the benchmark.
        experiment_names (list[str] | None, optional): List of names of experiments to be loaded, if None, all experiments in `savedir` will be loaded. Defaults to None.
        savedir (str | pathlib.Path, optional): Directory of the experiment. Defaults to DEFAULT_RESULT_DIR.
        return_type (tp.Literal[&quot;Artifact&quot;, &quot;Experiment&quot;, &quot;Table&quot;], optional): Type of the returned object. Defaults to "Experiment".
        index_col (int | list[int] | None, optional): The column(s) to set as the index(MultiIndex) of the returned Table.. Defaults to None.

    Raises:
        FileNotFoundError: If `name_or_dir` is not found in the `savedir` directory.
        ValueError: If `return_type` is not one of "Artifact", "Experiment", or "Table".

    Returns:
        Experiment | Artifact | Table: The loaded artifact, experiment, or table.
    """
    from jijbench.experiment.experiment import Experiment

    name_or_dir = (
        name_or_dir
        if isinstance(name_or_dir, pathlib.Path)
        else pathlib.Path(name_or_dir)
    )

    if name_or_dir.exists():
        savedir = name_or_dir
    else:
        savedir /= name_or_dir

    if not savedir.exists():
        raise FileNotFoundError(f"{name_or_dir} is not found.")

    if return_type == "Artifact":
        with open(f"{savedir}/artifact.dill", "rb") as f:
            return dill.load(f)
    elif return_type == "Experiment":
        if experiment_names is None:
            dirs = savedir.iterdir()
        else:
            dirs = [savedir / name for name in experiment_names]
        experiments = []
        for p in dirs:
            if p.is_dir():
                a = load("artifact", savedir=p, return_type="Artifact")
                t = load("table", savedir=p, return_type="Table", index_col=index_col)
                experiments.append(Experiment((a, t), p.name))
        concat: Concat[Experiment] = Concat()
        return concat(experiments)
    elif return_type == "Table":
        with open(f"{savedir}/table.dill", "rb") as f:
            return dill.load(f)
    else:
        raise ValueError("Argument return type must be Artifact, Experiment or Table.")



from pkgutil import extend_path

__path__ = extend_path(__path__, __name__)


from jijbench.functions.concat import Concat
from jijbench.functions.factory import ArtifactFactory, RecordFactory, TableFactory
from jijbench.functions.math import Max, Mean, Min, Std

__all__ = [
    "Concat",
    "ArtifactFactory",
    "RecordFactory",
    "TableFactory",
    "Max",
    "Mean",
    "Min",
    "Std",
]



from __future__ import annotations

import typing as tp
import warnings

import numpy as np

from jijbench.elements.base import Number
from jijbench.functions.concat import Concat
from jijbench.node.base import FunctionNode
from jijbench.solver.jijzept import SampleSet


def _is_success_list(sampleset: SampleSet, opt_value: int | float) -> list[bool]:
    is_feas = _is_feasible_list(sampleset)
    objective = np.array(sampleset.evaluation.objective)
    return list((objective <= opt_value) & is_feas)


def _is_feasible_list(sampleset: SampleSet) -> list[bool]:
    if sampleset.evaluation.constraint_violations is None:
        raise ValueError(
            "The value of sampleset.evaluation.constraint_violations is None. This SampleSet object is not evaluated or has not constraints."
        )
    else:
        constraint_violations = np.array(
            list(sampleset.evaluation.constraint_violations.values())
        )
        return constraint_violations.sum(axis=0) == 0


class Metrics(FunctionNode[SampleSet, Number]):
    """A base class for metrics."""

    def __call__(self, inputs: list[SampleSet], **kwargs: tp.Any) -> Number:
        """Calculate the metric.

        Args:
            inputs (list[SampleSet]): A list of SampleSet objects.
        """

        def _generate_warning_msg(metrics):
            return f'{self.__class__.__name__} cannot be calculated because "{metrics}" is not stored in table of Experiment object.'

        concat: Concat[SampleSet] = Concat()
        node = concat(inputs, "sampleset")

        objective = node.evaluation.objective
        if objective is None:
            raise ValueError("This SampleSet object is not evaluated.")
        if np.isnan(objective).any():
            warnings.warn(_generate_warning_msg("objective"))

        time = node.time.solve
        if time is None:
            warnings.warn(_generate_warning_msg("execution_time"))
        else:
            if time.solve is None:
                warnings.warn(_generate_warning_msg("execution_time"))

        num_occurrences = node.record.num_occurrences
        if np.isnan(num_occurrences).any():
            warnings.warn(_generate_warning_msg("num_occurrences"))

        return super().__call__([node], **kwargs)


class TimeToSolution(Metrics):
    """Time to solution.
    The time to solution is defined as the time to obtain a solution with probability `pr`.
    The solution is defined as the solution that is feasible and less than or equal to `opt_value`.
    There are three types of time to solution:
        - optimal: The solution is the optimal solution.
        - feasible: The solution is feasible.
        - derived: The solution is the best solution among the obtained solutions.
    """

    @tp.overload
    def __call__(
        self,
        inputs: list[SampleSet],
        pr: float,
        *,
        opt_value: int | float,
        base: tp.Literal["optimal"] = "optimal",
    ) -> Number:
        ...

    @tp.overload
    def __call__(
        self, inputs: list[SampleSet], pr: float, *, base: tp.Literal["feasible"]
    ) -> Number:
        ...

    @tp.overload
    def __call__(
        self, inputs: list[SampleSet], pr: float, *, base: tp.Literal["derived"]
    ) -> Number:
        ...

    def __call__(
        self,
        inputs: list[SampleSet],
        pr: float,
        *,
        opt_value: int | float | None = None,
        base: tp.Literal["optimal", "feasible", "derived"] = "optimal",
    ) -> Number:
        """Calculate time to solution.

        Args:
            inputs (list[SampleSet]): A list of SampleSet objects.
            pr (float): The probability of obtaining a solution.
            opt_value (int | float): The optimal value.
            base (str): The type of time to solution.

        Returns:
            Number: The time to solution.
        """
        return super().__call__(inputs, pr=pr, opt_value=opt_value, base=base)

    @tp.overload
    def operate(
        self,
        inputs: list[SampleSet],
        pr: float,
        *,
        opt_value: int | float,
        base: tp.Literal["optimal"] = "optimal",
    ) -> Number:
        ...

    @tp.overload
    def operate(
        self, inputs: list[SampleSet], pr: float, *, base: tp.Literal["feasible"]
    ) -> Number:
        ...

    @tp.overload
    def operate(
        self, inputs: list[SampleSet], pr: float, *, base: tp.Literal["derived"]
    ) -> Number:
        ...

    def operate(
        self,
        inputs: list[SampleSet],
        pr: float,
        *,
        opt_value: int | float | None = None,
        base: tp.Literal["optimal", "feasible", "derived"] = "optimal",
    ) -> Number:
        """Calculate time to solution.

        Args:
            inputs (list[SampleSet]): A list of SampleSet objects.
            pr (float): The probability of obtaining a solution.
            opt_value (int | float): The optimal value.
            base (str): The type of time to solution.

        Returns:
            Number: The time to solution.
        """

        node = inputs[0]

        num_occurrences = np.array(node.record.num_occurrences)
        if sum(num_occurrences) == 1:
            warnings.warn("num_reads = 1; should be increased to measure TTS")

        if base == "optimal":
            f_ps = SuccessProbability()
            ps = f_ps([node], opt_value=opt_value or np.nan)
        elif base == "feasible":
            f_ps = FeasibleRate()
            ps = f_ps([node])
        elif base == "derived":
            f_ps = SuccessProbability()
            is_feas = _is_feasible_list(node)
            if any(is_feas):
                opt_value = np.array(node.evaluation.objective)[is_feas].min()
                ps = f_ps([node], opt_value=opt_value or np.nan)
            else:
                ps = f_ps([node], opt_value=np.inf)

        time = np.nan if node.time.solve is None else node.time.solve.solve
        time = np.nan if time is None else time

        if ps.data == 1:
            data = 0.0
        elif ps:
            data = np.log(1 - pr) / np.log(1 - ps.data) * time
        else:
            data = np.inf
        return Number(data, f"TTS[{base}]")


class SuccessProbability(Metrics):
    """Success probability.
    The success probability is defined as the probability of obtaining a solution less than or equal to `opt_value`.
    """

    def __call__(self, inputs: list[SampleSet], opt_value: int | float) -> Number:
        """Calculate success probability.

        Args:
            inputs (list[SampleSet]): A list of SampleSet objects.
            opt_value (int | float): The optimal value.

        Returns:
            Number: The success probability.
        """
        return super().__call__(inputs, opt_value=opt_value)

    def operate(self, inputs: list[SampleSet], opt_value: int | float) -> Number:
        """Calculate success probability.

        Args:
            inputs (list[SampleSet]): A list of SampleSet objects.
            opt_value (int | float): The optimal value.

        Returns:
            Number: The success probability.
        """
        node = inputs[0]
        num_occurrences = np.array(node.record.num_occurrences)
        num_success = sum(_is_success_list(node, opt_value) * num_occurrences)
        data = num_success / sum(num_occurrences)
        return Number(data, "success_probability")


class FeasibleRate(Metrics):
    """Feasible rate.
    The feasible rate is defined as the probability of obtaining a feasible solution.
    """

    def __call__(self, inputs: list[SampleSet]) -> Number:
        """Calculate feasible rate.

        Args:
            inputs (list[SampleSet]): A list of SampleSet objects.

        Returns:
            Number: The feasible rate.

        """
        return super().__call__(inputs)

    def operate(self, inputs: list[SampleSet]) -> Number:
        node = inputs[0]
        num_occurrences = np.array(node.record.num_occurrences)
        num_feasible = sum(_is_feasible_list(node) * num_occurrences)
        data = num_feasible / sum(num_occurrences)
        return Number(data, "feasible_rate")


class ResidualEnergy(Metrics):
    """Residual energy.
    The residual energy is defined as the mean of the objective function minus the optimal value.
    """

    def __call__(self, inputs: list[SampleSet], opt_value: int | float) -> Number:
        return super().__call__(inputs, opt_value=opt_value)

    def operate(self, inputs: list[SampleSet], opt_value: int | float) -> Number:
        node = inputs[0]

        is_feas = np.array(_is_feasible_list(node))
        if all(~is_feas):
            data = np.nan
        else:
            num_occurrences = np.array(node.record.num_occurrences)
            objective = np.array(node.evaluation.objective) * is_feas * num_occurrences
            mean = objective.sum() / (is_feas * num_occurrences).sum()
            data = float(mean - opt_value)
        return Number(data, "residual_energy")



from abc import ABCMeta, abstractmethod


class Figure(metaclass=ABCMeta):
    # Display a figure using matplotlib
    @abstractmethod
    def show(self):
        raise NotImplementedError

    # Returns the matplotlib figure and axes created in show method.
    @property
    @abstractmethod
    def fig_ax(self):
        raise NotImplementedError



from __future__ import annotations

import datetime
from dataclasses import dataclass, field

import pandas as pd

from jijbench.node.base import DataNode
from jijbench.typing import DateTypes


@dataclass
class Date(DataNode[DateTypes]):
    """A class for date information.

    This class has date information in the `data` attribute.
    The data can be stored as a string, datetime.datetime object or pandas.Timestamp object.
    If the data is stored as a string, the class will try to convert it to pandas.Timestamp.

    Attributes:
        data (DateTypes): The date information stored in the node.
        name (str): The name of the node.
    """

    data: DateTypes = field(default_factory=pd.Timestamp.now)
    name: str = "timestamp"

    def __post_init__(self) -> None:
        if isinstance(self.data, str):
            self.data = pd.Timestamp(self.data)

        if isinstance(self.data, datetime.datetime):
            self.data = pd.Timestamp(self.data)

    @classmethod
    def validate_data(cls, data: DateTypes) -> DateTypes:
        """Validate data attribute and make sure it's a string, datetime.datetime object or pandas.Timestamp object.

        Raises:
            ValueError: If `data` attribute is a string and not a valid date string.

        Returns:
            DateTypes: The validated date data.
        """
        data = cls._validate_dtype(data, (str, datetime.datetime, pd.Timestamp))
        if isinstance(data, str):
            try:
                pd.Timestamp(data)
            except Exception:
                raise ValueError(f"Date string '{data}' is invalid for data attribute.")
        return data



from __future__ import annotations

import typing as tp

import numpy as np

from jijbench.elements.base import Number
from jijbench.node.base import FunctionNode

if tp.TYPE_CHECKING:
    from jijbench.elements.array import Array


class Min(FunctionNode["Array", Number]):
    """Calculate the minimum value of an input `Array`.

    The `Min` class is a subclass of `FunctionNode` that calculates the minimum value of an input `Array`
    and returns the result as a `Number` object.

    Attributes:
        inputs (List[Array]): A list of `Array` objects to operate.
        name (str): A name for the node.
    """

    def operate(self, inputs: list[Array]) -> Number:
        """Calculate the minimum value of the input `Array`.

        Args:
            inputs (List[Array]): A list of `Array` objects to operate.

        Returns:
            Number: The result of the calculation as a `Number` object.
        """
        return _operate_array(inputs, np.min, "min")


class Max(FunctionNode["Array", Number]):
    """Calculate the maximum value of an input `Array`.

    The `Max` class is a subclass of `FunctionNode` that calculates the maximum value of an input `Array`
    and returns the result as a `Number` object.

    Attributes:
        inputs (List[Array]): A list of `Array` objects to operate.
        name (str): A name for the node.
    """

    def operate(self, inputs: list[Array]) -> Number:
        """Calculate the maximum value of the input `Array`.

        Args:
            inputs (List[Array]): A list of `Array` objects to operate.

        Returns:
            Number: The result of the calculation as a `Number` object.
        """
        return _operate_array(inputs, np.max, "max")


class Mean(FunctionNode["Array", Number]):
    """Calculate the mean value of an input `Array`.

    The `Mean` class is a subclass of `FunctionNode` that calculates the mean value of an input `Array`
    and returns the result as a `Number` object.

    Attributes:
        inputs (List[Array]): A list of `Array` objects to operate.
        name (str): A name for the node.
    """

    def operate(self, inputs: list[Array]) -> Number:
        """Calculate the mean value of the input `Array`.

        Args:
            inputs (List[Array]): A list of `Array` objects to operate.

        Returns:
            Number: The result of the calculation as a `Number` object.
        """
        return _operate_array(inputs, np.mean)


class Std(FunctionNode["Array", Number]):
    """Calculate the standard deviation of an input `Array`.

    The `Std` class is a subclass of `FunctionNode` that calculates the standard deviation of an input `Array`
    and returns the result as a `Number` object.

    Attributes:
        inputs (List[Array]): A list of `Array` objects to operate.
        name (str): A name for the node.
    """

    def operate(self, inputs: list[Array]) -> Number:
        """Calculate the standard deviation of the input `Array`.

        Args:
            inputs (List[Array]): A list of `Array` objects to operate.

        Returns:
            Number: The result of the calculation as a `Number` object.
        """
        return _operate_array(inputs, np.std)


def _operate_array(
    inputs: list[Array], f: tp.Callable, f_name: str | None = None
) -> Number:
    data = f(inputs[0].data)
    if "int" in str(data.dtype):
        data = int(data)
    else:
        data = float(data)

    if f_name is None:
        f_name = f.__name__
    name = f"{inputs[0].name}_{f_name}"
    return Number(data, name)
